const grpc = require('grpc')
const Promise = require('bluebird')
const typedError = require('error/typed')
const map = require('ramda/src/map')
const pipe = require('ramda/src/pipe')
const path = require('path')
const keys = require('ramda/src/keys')
const isEmpty = require('ramda/src/isEmpty')
const fromPairs = require('ramda/src/fromPairs')
const t = require('tcomb')

const {implementationError, connectionError,} = require('../errors')

const grpcBadRequestError = typedError({
  type: 'grpcBadRequest',
  message: 'gRPC bad request',
})

const grpcNotFoundError = typedError({
  type: 'grpcNotFound',
  message: 'gRPC resource not found',
})

const grpcImplementationError = typedError({
  type: 'grpcImplementationError',
  message: 'gRPC internal error',
})

const DEFAULTS = {
  partition: 'cache',
  serviceUri: 'localhost:9000',
}

const GRPC_CODE2ERROR_TYPE = {
  3: grpcBadRequestError,
  5: grpcNotFoundError,
}

const proto = grpc.load(path.join(__dirname, './gunCachingService.proto')).gunCachingService

const createService = (settings, logger) => {
  t.Object(settings, ['settings',])
  t.Object(logger, ['logger',])

  const client = new proto.GunCachingService(settings.serviceUri,
    grpc.credentials.createSsl(
      Buffer.from(settings.grpcTlsCa), Buffer.from(settings.grpcTlsClientKey),
      Buffer.from(settings.grpcTlsClientCert)
    )
  )

  // XXX: This isn't documented anywhere, but it's the best I could find
  const methodNames = keys(proto.GunCachingService.service)
  return pipe(
    map((name) => {
      return [name, async (message) => {
        const promisifiedMethod = Promise.promisify(client[name], {context: client,})
        let attempts = 0
        const INTERVAL = 1000
        while (attempts < 3) {
          try {
            return await promisifiedMethod(message)
          }
          catch (error) {
            if (error.code === grpc.status.UNAVAILABLE) {
              ++attempts
              await Promise.delay(INTERVAL)
            }
            else {
              const errorType = GRPC_CODE2ERROR_TYPE[error.code] || grpcImplementationError
              const params = {
                message: (error.details || '').trim() != null ? error.details : error.message,
              }
              if (error.metadata != null) {
                const data = error.metadata.getMap()
                if (!isEmpty(keys(data))) {
                  params.data = data
                }
              }
              const translatedError = errorType(params)
              throw translatedError
            }
          }
        }

        throw connectionError({
          message: `Couldn't connect to GUN caching service`,
        })
      },]
    }, ),
    fromPairs
  )(methodNames)
}

const generatePath = (settings, id) => {
  t.Object(settings, ['settings',])
  t.String(settings.partition, ['settings', 'partition',])
  t.String(id, ['id',])
  return [settings.partition, id,]
}

exports = module.exports = (options) => {
  const settings = {
    ...DEFAULTS,
    ...(options || {}),
  }
  t.String(settings.serviceUri, ['settings', 'serviceUri',])
  t.String(settings.grpcTlsCa, ['settings', 'grpcTlsCa',])
  t.String(settings.grpcTlsClientKey, ['settings', 'grpcTlsClientKey',])
  t.String(settings.grpcTlsClientCert, ['settings', 'grpcTlsClientCert',])

  const state = {
    settings,
    service: null,
  }

  const start = (logger) => {
    if (state.service == null) {
      state.service = createService(state.settings, logger)
    }
  }

  const stop = () => {
    state.service = null
  }

  const get = async (id) => {
    t.String(id, ['id',])

    if (state.service == null) {
      throw implementationError({message: 'Connection not started',})
    }

    const path = generatePath(state.settings, id)
    const result = await state.service.getEntry({path,})
    result.ttl = Number(result.ttl)
    result.stored = Number(result.stored)
    return result
  }

  const set = async (id, item, ttl) => {
    t.String(['id',])
    t.Number(ttl, ['ttl',])

    if (state.service == null) {
      throw implementationError({message: 'Connection not started',})
    }

    const path = generatePath(state.settings, id)
    await state.service.setEntry({path, item, ttl,})
  }

  const drop = async (id) => {
    if (state.service == null) {
      throw implementationError({message: 'Connection not started',})
    }

    const path = generatePath(state.settings, id)
    await state.service.deleteEntry({path,})
  }

  return {
    start,
    stop,
    get,
    set,
    drop,
  }
}
