// A cache manager plugin for Hapi
const t = require('tcomb')
const pick = require('ramda/src/pick')
const {DateTime,} = require('luxon')
const omit = require('ramda/src/omit')
const concat = require('ramda/src/concat')

const pkg = require('../package.json')
const {argumentError,} = require('./errors')

const DEFAULTS = {
  engine: 'local',
  expiresIn: 60 * 1000,
  generateTimeout: 5 * 1000,
}

const CACHE_SETTING_KEYS = ['expiresIn', 'generateTimeout',]

// Register Hapi.js plugin
const register = (server, options) => {
  t.Object(server, ['server',])
  t.Function(server.decorate, ['server', 'decorate',])

  const settings = {
    ...DEFAULTS,
    ...(options || {}),
  }

  const logger = server.logger()
  logger.debug({settings,}, `Registering cache manager plugin, engine: ${settings.engine}`)

  let engine
  if (settings.engine === 'local') {
    engine = require('./gunEngine/local')
  } else if (settings.engine === 'service') {
    engine = require('./gunEngine/service')
  } else {
    throw argumentError({message: `Unrecognized engine ${settings.engine}`,})
  }
  const client = engine(omit(concat(['engine',], CACHE_SETTING_KEYS), options))

  const cacheManager = {
    policies: new Map(),
    client,
    ...pick(CACHE_SETTING_KEYS, settings),
  }

  server.ext({
    type: 'onPreStart',
    method: async () => {
      logger.debug(`Starting cache`)
      await client.start(logger)
    },
  })
  server.ext({
    type: 'onPostStop',
    method: async () => {
      logger.debug(`Stopping cache`)
      await client.stop(logger)
    },
  })

  server.decorate('server', 'createCachePolicy', (options) => {
    return createCachePolicy(cacheManager, options, server)
  })
}

const createCachePolicy = (manager, options, server) => {
  t.Object(manager, ['manager',])
  t.Object(options, ['options',])
  t.String(options.name, ['options', 'name',])
  t.Function(options.generateFunc, ['options', 'generateFunc',])
  t.Object(server, ['server',])
  t.Function(server.logger, ['server', 'logger',])

  const logger = server.logger()

  if (manager.policies.has(options.name)) {
    throw argumentError({message: `Policy ${options.name} already exists`,})
  }

  logger.debug(`Creating cache policy ${options.name}`)
  const policy = {
    name: options.name,
    pendings: new Map(),
    client: manager.client,
    ...pick(['generateFunc',], options),
    ...pick(CACHE_SETTING_KEYS, manager),
  }
  policy.get = async (key, logger) => {
    return await getFromCache(policy, key, logger)
  }
  policy.drop = async (key, logger) => {
    await policy.client.drop(getId(policy, key), logger)
  }

  manager.policies.set(options.name, policy)

  return policy
}

const getId = (cachePolicy, key) => {
  t.Object(cachePolicy, ['cachePolicy',])
  t.Object(key, ['key',])
  return `${cachePolicy.name}/${key.id}`
}

const generateValue = async (cachePolicy, key, logger) => {
  t.Object(cachePolicy, ['cachePolicy',])
  t.Function(cachePolicy.generateFunc, ['cachePolicy', 'generateFunc',])
  t.Object(cachePolicy.client, ['cachePolicy', 'client',])
  t.Object(key, ['key',])
  t.String(key.id, ['key', 'id',])
  t.Object(logger, ['logger',])

  const value = await cachePolicy.generateFunc(key, logger)

  const ttl = cachePolicy.expiresIn
  const id = getId(cachePolicy, key)
  logger.debug(`Setting value in client`, {id, ttl,})
  const valueJson = JSON.stringify(value)
  await cachePolicy.client.set(id, valueJson, ttl, logger)
  return value
}

const getFromCache = async (cachePolicy, key, logger) => {
  t.Object(cachePolicy, ['cachePolicy',])
  t.Object(key, ['key',])
  t.String(key.id, ['key', 'id',])
  t.Object(logger, ['logger',])

  const id = getId(cachePolicy, key)
  logger.debug(`Trying to get entry ${id} from cache`)
  const pending = cachePolicy.pendings.get(id)
  if (pending != null) {
    logger.debug(`Awaiting pending promise for cache entry ${id}`)
    return await pending
  } else {
    logger.debug({cachePolicy: cachePolicy.name, id,}, `Getting value for entry ${id} from cache`)
    const cached = await cachePolicy.client.get(id, logger)
    if (cached != null) {
      logger.debug(`Entry ${id} retrieved from cache`)
      t.Number(cached.stored, ['cached', 'stored',])
      t.Number(cached.ttl, ['cached', 'ttl',])
      if (cached.ttl < (DateTime.utc().toMillis() - cached.stored)) {
        logger.debug(`Cache entry ${id} has expired, dropping it`)
        await cachePolicy.client.drop(id, logger)
      } else {
        logger.debug(`Returning cache entry ${id} as it's not yet expired`)
        return JSON.parse(cached.item)
      }
    }

    logger.debug(`Generating value for cache entry ${id} - adding to pending promises`)
    const promise = generateValue(cachePolicy, key, logger)
    cachePolicy.pendings.set(id, promise)
    try {
      return await promise
    } finally {
      logger.debug(`Deleting pending promise for cache entry ${id}`)
      cachePolicy.pendings.delete(id)
    }
  }
}

module.exports = {
  plugin: {
    name: 'cacheManager',
    version: pkg.version,
    register,
  },
}
